#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2013 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides some basic usable widgets based on TaurusReadWriteSwitcher
"""

__all__ = ["TaurusLabelEditRW", "TaurusLabelEditRW"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.qt.qtgui.display import TaurusLabel, TaurusLed
from taurus.qt.qtgui.input import TaurusValueLineEdit, TaurusValueCheckBox
from abstractswitcher import TaurusReadWriteSwitcher


class TaurusLabelEditRW(TaurusReadWriteSwitcher):
    '''A Switcher combining a TaurusLabel and a TaurusValueLineEdit'''
    readWClass = TaurusLabel
    writeWClass = TaurusValueLineEdit


class TaurusBoolRW(TaurusReadWriteSwitcher):
    '''A Switcher combining a TaurusLed and a TaurusValueCheckBox'''
    readWClass = TaurusLed
    writeWClass = TaurusValueCheckBox

    def setWriteWidget(self, widget):
        widget.setShowText(False)
        TaurusReadWriteSwitcher.setWriteWidget(self, widget)


def _demo():
    '''demo of integrability in a form'''
    import sys
    from taurus.qt.qtgui.panel import TaurusForm
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication()

    f = TaurusForm()
    f.model = ['sys/tg_test/1/long_scalar', 'sys/tg_test/1/long_scalar',
               'sys/tg_test/1/boolean_scalar', 'sys/tg_test/1/boolean_scalar']

    f[0].setReadWidgetClass(TaurusLabelEditRW)
    f[0].setWriteWidgetClass(None)
    f[2].setReadWidgetClass(TaurusBoolRW)
    f[2].setWriteWidgetClass(None)

    f.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    _demo()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
propertyfile.py:

A Python replacement for java.util.Properties class
This is modelled as closely as possible to the Java original.

Adapted from http://code.activestate.com/recipes/496795/
Created - Anand B Pillai <abpillai@gmail.com>
Modified - Tiago Coutinho
"""

__all__ = ["Properties"]

__docformat__ = "restructuredtext"

import sys
import os
import re
import time


class Properties(object):
    """ A Python replacement for java.util.Properties """

    def __init__(self, props=None):

        # Note: We don't take a default properties object
        # as argument yet

        # Dictionary of properties.
        self._props = {}
        # Dictionary of properties with 'pristine' keys
        # This is used for dumping the properties to a file
        # using the 'store' method
        self._origprops = {}

        # Dictionary mapping keys from property
        # dictionary to pristine dictionary
        self._keymap = {}

        self.othercharre = re.compile(r'(?<!\\)(\s*\=)|(?<!\\)(\s*\:)')
        self.othercharre2 = re.compile(r'(\s*\=)|(\s*\:)')
        self.bspacere = re.compile(r'\\(?!\s$)')

    def __str__(self):
        s = '{'
        for key, value in self._props.items():
            s = ''.join((s, key, '=', value, ', '))

        s = ''.join((s[:-2], '}'))
        return s

    def __parse(self, lines):
        """ Parse a list of lines and create
        an internal property dictionary """

        # Every line in the file must consist of either a comment
        # or a key-value pair. A key-value pair is a line consisting
        # of a key which is a combination of non-white space characters
        # The separator character between key-value pairs is a '=',
        # ':' or a whitespace character not including the newline.
        # If the '=' or ':' characters are found, in the line, even
        # keys containing whitespace chars are allowed.

        # A line with only a key according to the rules above is also
        # fine. In such case, the value is considered as the empty string.
        # In order to include characters '=' or ':' in a key or value,
        # they have to be properly escaped using the backslash character.

        # Some examples of valid key-value pairs:
        #
        # key     value
        # key=value
        # key:value
        # key     value1,value2,value3
        # key     value1,value2,value3 \
        #         value4, value5
        # key
        # This key= this value
        # key = value1 value2 value3

        # Any line that starts with a '#' is considerered a comment
        # and skipped. Also any trailing or preceding whitespaces
        # are removed from the key/value.

        # This is a line parser. It parses the
        # contents like by line.

        lineno = 0
        i = iter(lines)

        for line in i:
            lineno += 1
            line = line.strip()
            # Skip null lines
            if not line:
                continue
            # Skip lines which are comments
            if line[0] == '#':
                continue
            # Some flags
            escaped = False
            # Position of first separation char
            sepidx = -1
            # A flag for performing wspace re check
            flag = 0
            # Check for valid space separation
            # First obtain the max index to which we
            # can search.
            m = self.othercharre.search(line)
            if m:
                first, last = m.span()
                start, end = 0, first
                flag = 1
                wspacere = re.compile(r'(?<![\\\=\:])(\s)')
            else:
                if self.othercharre2.search(line):
                    # Check if either '=' or ':' is present
                    # in the line. If they are then it means
                    # they are preceded by a backslash.

                    # This means, we need to modify the
                    # wspacere a bit, not to look for
                    # : or = characters.
                    wspacere = re.compile(r'(?<![\\])(\s)')
                start, end = 0, len(line)

            m2 = wspacere.search(line, start, end)
            if m2:
                # print 'Space match=>',line
                # Means we need to split by space.
                first, last = m2.span()
                sepidx = first
            elif m:
                # print 'Other match=>',line
                # No matching wspace char found, need
                # to split by either '=' or ':'
                first, last = m.span()
                sepidx = last - 1
                # print line[sepidx]

            # If the last character is a backslash
            # it has to be preceded by a space in which
            # case the next line is read as part of the
            # same property
            while line[-1] == '\\':
                # Read next line
                nextline = i.next()
                nextline = nextline.strip()
                lineno += 1
                # This line will become part of the value
                line = line[:-1] + nextline

            # Now split to key,value according to separation char
            if sepidx != -1:
                key, value = line[:sepidx], line[sepidx + 1:]
            else:
                key, value = line, ''

            self.processPair(key, value)

    def processPair(self, key, value):
        """ Process a (key, value) pair """

        oldkey = key
        oldvalue = value

        # Create key intelligently
        keyparts = self.bspacere.split(key)
        # print keyparts

        strippable = False
        lastpart = keyparts[-1]

        if lastpart.find('\\ ') != -1:
            keyparts[-1] = lastpart.replace('\\', '')

        # If no backspace is found at the end, but empty
        # space is found, strip it
        elif lastpart and lastpart[-1] == ' ':
            strippable = True

        key = ''.join(keyparts)
        if strippable:
            key = key.strip()
            oldkey = oldkey.strip()

        oldvalue = self.unescape(oldvalue)
        value = self.unescape(value)

        self._props[key] = value.strip()

        # Check if an entry exists in pristine keys
        if self._keymap.has_key(key):
            oldkey = self._keymap.get(key)
            self._origprops[oldkey] = oldvalue.strip()
        else:
            self._origprops[oldkey] = oldvalue.strip()
            # Store entry in keymap
            self._keymap[key] = oldkey

    def escape(self, value):

        # Java escapes the '=' and ':' in the value
        # string with backslashes in the store method.
        # So let us do the same.
        newvalue = value.replace(':', '\:')
        newvalue = newvalue.replace('=', '\=')

        return newvalue

    def unescape(self, value):

        # Reverse of escape
        newvalue = value.replace('\:', ':')
        newvalue = newvalue.replace('\=', '=')

        return newvalue

    def load(self, stream):
        """ Load properties from an open file stream """

        # For the time being only accept file input streams
        if type(stream) is not file:
            raise TypeError, 'Argument should be a file object!'
        # Check for the opened mode
        if stream.mode != 'r':
            raise ValueError, 'Stream should be opened in read-only mode!'

        try:
            lines = stream.readlines()
            self.__parse(lines)
        except IOError, e:
            raise

    def getProperty(self, key):
        """ Return a property for the given key """

        return self._props.get(key, '')

    def setProperty(self, key, value):
        """ Set the property for the given key """

        if type(key) is str and type(value) is str:
            self.processPair(key, value)
        else:
            raise TypeError, 'both key and value should be strings!'

    def propertyNames(self):
        """ Return an iterator over all the keys of the property
        dictionary, i.e the names of the properties """

        return self._props.keys()

    def list(self, out=sys.stdout):
        """ Prints a listing of the properties to the
        stream 'out' which defaults to the standard output """

        out.write('-- listing properties --\n')
        for key, value in self._props.items():
            out.write(''.join((key, '=', value, '\n')))

    def store(self, out, header=""):
        """ Write the properties list to the stream 'out' along
        with the optional 'header' """

        if out.mode[0] != 'w':
            raise ValueError, 'Steam should be opened in write mode!'

        try:
            out.write(''.join(('#', header, '\n')))
            # Write timestamp
            tstamp = time.strftime('%a %b %d %H:%M:%S %Z %Y', time.localtime())
            out.write(''.join(('#', tstamp, '\n')))
            # Write properties from the pristine dictionary
            for prop, val in self._origprops.items():
                out.write(''.join((prop, '=', self.escape(val), '\n')))

            out.close()
        except IOError, e:
            raise

    def getPropertyDict(self):
        return self._props

    def __getitem__(self, name):
        """ To support direct dictionary like access """

        return self.getProperty(name)

    def __setitem__(self, name, value):
        """ To support direct dictionary like access """

        self.setProperty(name, value)

    def __getattr__(self, name):
        """ For attributes not found in self, redirect
        to the properties dictionary """

        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self._props, name):
                return getattr(self._props, name)

if __name__ == "__main__":
    p = Properties()
    p.load(open('test2.properties'))
    p.list()
    p['hello'] = "no"

#    print p
#    print p.items()
#    print p['name3']
#    p['name3'] = 'changed = value'
#    print p['name3']
#    p['new key'] = 'new value'
    p.store(open('test2.properties', 'w'))



#!/usr/bin/env python

import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.plot import TaurusTrend

app = Qt.QApplication(sys.argv)
##########################
# BEGIN EXAMPLE CODE
##########################

panel = TaurusTrend()
model = ['sys/taurustest/1/position']
panel.setXIsTime(True)  # to show the x values as time
panel.setModel(model)

########################
# END EXAMPLE CODE
########################
panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of basic taurus widgets based on QLineEdit"""

__all__ = ["TaurusValueLineEdit"]

__docformat__ = 'restructuredtext'

import sys
import numpy
from taurus.external.qt import Qt
from taurus.external.pint import Quantity
from taurus.qt.qtgui.base import TaurusBaseWritableWidget
from taurus.qt.qtgui.util import PintValidator
from taurus.core import DataType, DataFormat, TaurusEventType


class TaurusValueLineEdit(Qt.QLineEdit, TaurusBaseWritableWidget):

    def __init__(self, qt_parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QLineEdit, qt_parent)
        self.call__init__(TaurusBaseWritableWidget,
                          name, designMode=designMode)
        self._enableWheelEvent = False

        self.setAlignment(Qt.Qt.AlignRight)
        self.setValidator(None)

        self.textChanged.connect(self.notifyValueChanged)
        self.returnPressed.connect(self.writeValue)
        self.valueChangedSignal.connect(self.updatePendingOperations)
        self.editingFinished.connect(self._onEditingFinished)

    def _updateValidator(self, value):
        '''This method sets a validator depending on the data type'''
        if isinstance(value.wvalue, Quantity):
            val = self.validator()
            if not isinstance(val, PintValidator):
                val = PintValidator(self)
                self.setValidator(val)
            attr = self.getModelObj()
            bottom, top = attr.range
            if bottom != val.bottom:
                val.setBottom(bottom)
            if top != val.top:
                val.setTop(top)
            units = value.wvalue.units
            if units != val.units:
                val.setUnits(units)

        # @TODO Other validators can be configured for other types (e.g. with string lengths, tango names,...)
        else:
            self.setValidator(None)
            self.debug("Validator disabled")

    def __decimalDigits(self, fmt):
        '''returns the number of decimal digits from a format string
        (or None if they are not defined)'''
        try:
            if fmt[-1].lower() in ['f', 'g'] and '.' in fmt:
                return int(fmt[:-1].split('.')[-1])
            else:
                return None
        except:
            return None

    def _onEditingFinished(self):
        '''slot for performing autoapply only when edition is finished'''
        if self._autoApply:
            self.writeValue()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWritableWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def notifyValueChanged(self, *args):
        '''reimplement to avoid autoapply on every partial edition'''
        self.emitValueChanged()

    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
            self._updateValidator(evt_value)
        TaurusBaseWritableWidget.handleEvent(
            self, evt_src, evt_type, evt_value)

    def updateStyle(self):
        TaurusBaseWritableWidget.updateStyle(self)

        value = self.getValue()
        if value is None:
            # invalid value
            color, weight = 'gray', 'normal'
        else:
            # check if there are pending operations
            if self.hasPendingOperations():
                color, weight = 'blue', 'bold'
            else:
                color, weight = 'black', 'normal'
            # also check alarms (if applicable)
            modelObj = self.getModelObj()
            if modelObj and modelObj.type in [DataType.Integer, DataType.Float]:
                min_, max_ = modelObj.alarms
                if value < min_ or value > max_:
                    color = 'orange'
        # apply style
        style = 'TaurusValueLineEdit {color: %s; font-weight: %s}' %\
                (color, weight)
        self.setStyleSheet(style)

    def wheelEvent(self, evt):
        if not self.getEnableWheelEvent() or Qt.QLineEdit.isReadOnly(self):
            return Qt.QLineEdit.wheelEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return Qt.QLineEdit.wheelEvent(self, evt)

        evt.accept()
        numDegrees = evt.delta() / 8
        numSteps = numDegrees / 15
        modifiers = evt.modifiers()
        if modifiers & Qt.Qt.ControlModifier:
            numSteps *= 10
        elif (modifiers & Qt.Qt.AltModifier) and model.type == DataType.Float:
            numSteps *= .1
        self._stepBy(numSteps)

    def keyPressEvent(self, evt):
        if evt.key() in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            Qt.QLineEdit.keyPressEvent(self, evt)
            evt.accept()
            return
        if Qt.QLineEdit.isReadOnly(self):
            return Qt.QLineEdit.keyPressEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return Qt.QLineEdit.keyPressEvent(self, evt)

        if evt.key() == Qt.Qt.Key_Up:
            numSteps = 1
        elif evt.key() == Qt.Qt.Key_Down:
            numSteps = -1
        else:
            return Qt.QLineEdit.keyPressEvent(self, evt)

        evt.accept()
        modifiers = evt.modifiers()
        if modifiers & Qt.Qt.ControlModifier:
            numSteps *= 10
        elif (modifiers & Qt.Qt.AltModifier) and model.type == DataType.Float:
            numSteps *= .1
        self._stepBy(numSteps)

    def _stepBy(self, v):
        value = self.getValue()
        self.setValue(value + Quantity(v, value.units))

    def setValue(self, v):
        model = self.getModelObj()
        if model is None:
            v_str = str(v)
        else:
            v_str = str(self.displayValue(v))
        v_str = v_str.strip()
        self.setText(v_str)

    def getValue(self):
        text = self.text()
        model_obj = self.getModelObj()
        if model_obj is None:
            return None
        val = self.validator()
        try:
            model_type = model_obj.type
            model_format = model_obj.data_format
            if model_type in [DataType.Integer, DataType.Float]:
                if val is None or \
                        val.validate(str(text), 0)[0] != val.Acceptable:
                    return None
                q = Quantity(text)
                # allow implicit units (assume wvalue.units implicitly)
                if q.dimensionless:
                    q = Quantity(q.magnitude, val.units)
                return q
            elif model_type == DataType.Boolean:
                if model_format == DataFormat._0D:
                    return bool(int(eval(text)))
                else:
                    return numpy.array(eval(text), dtype=int).astype(bool)
            elif model_type == DataType.String:
                if model_format == DataFormat._0D:
                    return str(text)
                else:
                    return numpy.array(eval(text), dtype=str).tolist()
            elif model_type == DataType.Bytes:
                return bytes(text)
            else:
                raise TypeError('Unsupported model type "%s"', model_type)
        except Exception, e:
            self.warning('Cannot return value for "%s". Reason: %r', text, e)
            return None

    def setEnableWheelEvent(self, b):
        self._enableWheelEvent = b

    def getEnableWheelEvent(self):
        return self._enableWheelEvent

    def resetEnableWheelEvent(self):
        self.setEnableWheelEvent(False)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWritableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:lineedit.png"
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseWritableWidget.getModel,
                            TaurusBaseWritableWidget.setModel,
                            TaurusBaseWritableWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getUseParentModel,
                                     TaurusBaseWritableWidget.setUseParentModel,
                                     TaurusBaseWritableWidget.resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getAutoApply,
                                TaurusBaseWritableWidget.setAutoApply,
                                TaurusBaseWritableWidget.resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getForcedApply,
                                  TaurusBaseWritableWidget.setForcedApply,
                                  TaurusBaseWritableWidget.resetForcedApply)

    enableWheelEvent = Qt.pyqtProperty("bool", getEnableWheelEvent,
                                       setEnableWheelEvent,
                                       resetEnableWheelEvent)


def main():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication()

    form = Qt.QWidget()
    layout = Qt.QVBoxLayout()
    form.setLayout(layout)
    for m in ('sys/tg_test/1/double_scalar',
              'sys/tg_test/1/double_scalar'
              ):
        w = TaurusValueLineEdit()
        w.setModel(m)
        layout.addWidget(w)
    form.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    sys.exit(main())



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.input import TaurusWheelEdit

app = Qt.QApplication(sys.argv)

panel = TaurusWidget()
layout = Qt.QHBoxLayout()
panel.setLayout(layout)

w1 = TaurusLabel()
w2 = TaurusLabel()
w3 = TaurusWheelEdit()
w4 = TaurusLabel()
layout.addWidget(w1)
layout.addWidget(w2)
layout.addWidget(w3)
layout.addWidget(w4)
w1.setUseParentModel(True)
w2.setUseParentModel(True)
w3.setUseParentModel(True)
w4.setUseParentModel(True)
panel.setModel('sys/taurustest/1')
w1.setModel('/position#label')
w2.setModel('/position')
w3.setModel('/position')
w4.setModel('/position#unit')

panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for Taurus Forms"""

from taurus.external import unittest
from taurus.qt.qtgui.test import GenericWidgetTestCase
from taurus.qt.qtgui.panel import TaurusForm, TaurusAttrForm


class TaurusFormTest(GenericWidgetTestCase, unittest.TestCase):

    '''
    Generic tests for TaurusForm widget.

    .. seealso: :class:`taurus.qt.qtgui.test.base.GenericWidgetTestCase`
    '''
    _klass = TaurusForm
    modelnames = [['sys/tg_test/1'],
                  ['sys/tg_test/1/wave'],
                  [],
                  '',
                  ['eval:1'],
                  None,
                  ['sys/tg_test/1/%s' % a for a in (
                   'short_scalar', 'double_array',
                   'uchar_image_ro', 'string_spectrum',
                   'no_value', 'throw_exception')],
                  [''],
                  'sys/tg_test/1,eval:1',
                  'sys/tg_test/1/short_image eval:rand(16)',
                  [None]
                  ]


class TaurusAttrFormTest(GenericWidgetTestCase, unittest.TestCase):

    '''
    Generic tests for TaurusAttrForm widget.

    .. seealso: :class:`taurus.qt.qtgui.test.base.GenericWidgetTestCase`
    '''
    _klass = TaurusAttrForm
    modelnames = ['sys/tg_test/1', None]


# if __name__ == "__main__":
#     unittest.main()
#    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TaurusFormTest)
#    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
This module provides objects to manage macro-related tasks. Its primary use is
to be used within a TaurusGui for managing panels for:
- setting preferences in the sardana control system for data I/O
- displaying results of macro executions, including creating/removing panels for
  plotting results of scans
- editing macros

.. note:: This module will be moved to sardana.taurus at some point.
"""

 # TODO: move to sardana.taurus

__all__ = ['MacroBroker', 'DynamicPlotManager']
__docformat__ = 'restructuredtext'

import datetime

from taurus.core.util.containers import CaselessList
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseComponent
from sardana.taurus.core.tango.sardana import PlotType
from sardana.taurus.core.tango.sardana.pool import getChannelConfigs


class ChannelFilter(object):

    def __init__(self, chlist):
        self.chlist = tuple(chlist)

    def __call__(self, x):
        return x in self.chlist


class DynamicPlotManager(Qt.QObject, TaurusBaseComponent):
    '''This is a manager of plots related to the execution of macros.
    It dynamically creates/removes plots according to the configuration made by
    an ExperimentConfiguration widget.

    Currently it supports only 1D scan trends (2D scans are only half-baked)

    To use it simply instantiate it and pass it a door name as a model. You may
    want to call :meth:`onExpConfChanged` to update the configuration being
    used.
    '''

    newShortMessage = Qt.pyqtSignal(str)

    def __init__(self, parent=None):
        Qt.QObject.__init__(self, parent)
        TaurusBaseComponent.__init__(self, self.__class__.__name__)

        self.__panels = {}

        self._trends1d = {}
        self._trends2d = {}

    def setModel(self, doorname):
        '''reimplemented from :meth:`TaurusBaseComponent`

        :param doorname: (str) device name corresponding to a Door device.
        '''
        TaurusBaseComponent.setModel(self, doorname)
        # self._onDoorChanged(doorname)
        if not doorname:
            return
        door = self.getModelObj()
        if not isinstance(door, Qt.QObject):
            msg = "Unexpected type (%s) for %s" % (repr(type(door)), doorname)
            Qt.QMessageBox.critical(
                self.parent(), 'Door connection error', msg)
            return

        self._checkJsonRecorder()

        # read the expconf
        expconf = door.getExperimentConfiguration()
        self.onExpConfChanged(expconf)

    def _checkJsonRecorder(self):
        '''Checks if JsonRecorder env var is set and offers to set it'''
        door = self.getModelObj()
        if 'JsonRecorder' not in door.getEnvironment():
            msg = ('JsonRecorder environment variable is not set, but it ' +
                   'is needed for displaying trend plots.\n' +
                   'Enable it globally for %s?') % door.fullname
            result = Qt.QMessageBox.question(self.parent(),
                                             'JsonRecorder not set', msg,
                                             Qt.QMessageBox.Yes | Qt.QMessageBox.No)
            if result == Qt.QMessageBox.Yes:
                door.putEnvironment('JsonRecorder', True)
                self.info('JsonRecorder Enabled for %s' % door.fullname)

    def onExpConfChanged(self, expconf):
        '''
        Slot to be called when experimental configuration changes. It should
        remove the temporary panels and create the new ones needed.

        :param expconf: (dict) An Experiment Description dictionary. See
                        :meth:`sardana.taurus.qt.qtcore.tango.sardana.
                        QDoor.getExperimentDescription`
                        for more details
        '''
        if expconf['ActiveMntGrp'] is None:
            return
        if expconf['ActiveMntGrp'] not in expconf['MntGrpConfigs'].keys():
            self.warning(
                "ActiveMntGrp '%s' is not defined" %
                expconf['ActiveMntGrp'])
            return
        mgconfig = expconf['MntGrpConfigs'][expconf['ActiveMntGrp']]
        channels = dict(getChannelConfigs(mgconfig, sort=False))

        # classify by type of plot:
        trends1d = {}
        trends2d = {}
        plots1d = {}
        images = {}

        for chname, chdata in channels.items():
            ptype = chdata['plot_type']
            if ptype == PlotType.No:
                continue
            elif ptype == PlotType.Spectrum:
                axes = tuple(chdata['plot_axes'])
                # TODO: get default value from the channel.
                ndim = chdata.get('ndim', 0)
                if ndim == 0:  # this is a trend
                    if axes in trends1d:
                        trends1d[axes].append(chname)
                    else:
                        trends1d[axes] = CaselessList([chname])
                elif ndim == 1:  # a 1D plot (e.g. a spectrum)
                    pass  # TODO: implement
                else:
                    self.warning('Cannot create plot for %s', chname)

            elif ptype == PlotType.Image:
                axes = tuple(chdata['plot_axes'])
                # TODO: get default value from the channel.
                ndim = chdata.get('ndim', 1)
                if ndim == 0:  # a mesh-like plot?
                    pass  # TODO implement
                elif ndim == 1:  # a 2D trend
                    if axes in trends2d:
                        trends2d[axes].append(chname)
                    else:
                        trends2d[axes] = CaselessList([chname])
                elif ndim == 2:  # a 2D plot (e.g. an image)
                    pass  # TODO: implement
                else:
                    self.warning('Cannot create plot for %s', chname)

        new1d, removed1d = self._updateTemporaryTrends1D(trends1d)
        self.newShortMessage.emit("Changed panels (%i new, %i removed)" % (len(new1d),
                                                           len(removed1d)))
#        self._updateTemporaryTrends2D(trends2d)

    def _updateTemporaryTrends1D(self, trends1d):
        '''adds necessary trend1D panels and removes no longer needed ones

        :param trends1d: (dict) A dict whose keys are tuples of axes and
                         whose values are list of model names to plot

        :returns: (tuple) two lists new,rm:new contains the names of the new
                  panels and rm contains the names of the removed panels
        '''
        from taurus.qt.qtgui.plot import TaurusTrend
        newpanels = []
        for axes, plotables in trends1d.items():
            if not axes:
                continue
            if axes not in self._trends1d:
                w = TaurusTrend()
                w.setXIsTime(False)
                w.setScanDoor(self.getModelObj().fullname)
                # TODO: use a standard key for <idx> and <mov>
                w.setScansXDataKey(axes[0])
                pname = u'Trend1D - %s' % ":".join(axes)
                panel = self.createPanel(w, pname, registerconfig=False,
                                         permanent=False)
                try:  # if the panel is a dockwidget, raise it
                    panel.raise_()
                except:
                    pass
                self._trends1d[axes] = pname
                newpanels.append(pname)

            widget = self.getPanelWidget(self._trends1d[axes])
            flt = ChannelFilter(plotables)
            widget.onScanPlotablesFilterChanged(flt)

        # remove trends that are no longer configured
        removedpanels = []
        olditems = list(self._trends1d.items())
        for axes, name in olditems:
            if axes not in trends1d:
                removedpanels.append(name)
                self.removePanel(name)
                self._trends1d.pop(axes)

        return newpanels, removedpanels

    def _updateTemporaryTrends2D(self, trends2d):
        '''adds necessary trend2D panels and removes no longer needed ones

        :param trends2d: (dict) A dict whose keys are tuples of axes and
                         whose values are list of model names to plot

        :returns: (tuple) two lists new,rm:new contains the names of the new
                  panels and rm contains the names of the removed panels

        ..note:: Not fully implemented yet
        '''
        try:
            from taurus.qt.qtgui.extra_guiqwt.taurustrend2d import \
                TaurusTrend2DDialog
            from taurus.qt.qtgui.extra_guiqwt.image import TaurusTrend2DScanItem
        except:
            self.info('guiqwt extension cannot be loaded. ' +
                      '2D Trends will not be created')
            raise
            return

        for axes, plotables in trends2d.items():
            for chname in plotables:
                pname = u'Trend2D - %s' % chname
                if pname in self._trends2d:
                    self._trends2d[pname].widget().trendItem.clearTrend()
                else:
                    axis = axes[0]
                    w = TaurusTrend2DDialog(stackMode='event')
                    plot = w.get_plot()
                    t2d = TaurusTrend2DScanItem(chname, axis,
                                                self.getModelObj().fullname)
                    plot.add_item(t2d)
                    self.createPanel(w, pname, registerconfig=False,
                                     permanent=False)
                    self._trends2d[(axes, chname)] = pname

    def createPanel(self, widget, name, **kwargs):
        '''Creates a "panel" from a widget. In this basic implementation this
        means that the widgets is shown as a non-modal top window

        :param widget: (QWidget) widget to be used for the panel
        :param name: (str) name of the panel. Must be unique.

        Note: for backawards compatibility, this implementation accepts
        arbitrary keyword arguments which are just ignored
        '''
        widget.setWindowTitle(name)
        widget.show()
        self.__panels[name] = widget

    def getPanelWidget(self, name):
        '''Returns the widget associated to a panel name

        :param name: (str) name of the panel. KeyError is raised if not found

        :return: (QWidget)
        '''
        return self.__panels[name]

    def removePanel(self, name):
        '''stop managing the given panel

        :param name: (str) name of the panel'''
        widget = self.__panels.pop(name)
        if hasattr(widget, 'setModel'):
            widget.setModel(None)
        widget.setParent(None)
        widget.close()

    def removePanels(self, names=None):
        '''removes panels.

        :param names: (seq<str>) names of the panels to be removed. If None is
                      given (default), all the panels are removed.
        '''
        if names is None:
            names = self._trends1d.values() + self._trends2d.values()
            # TODO: do the same for other temporary panels
        for pname in names:
            self.removePanel(pname)


class MacroBroker(DynamicPlotManager):
    '''A manager of all macro-related panels of a TaurusGui.

    It creates, destroys and manages connections for the following objects:

        - Macro Configuration dialog
        - Experiment Configuration panel
        - Macro Executor panel
        - Sequencer panel
        - Macro description viewer
        - Door output, result and debug panels
        - Macro editor
        - Macro "panic" button (to abort macros)
        - Dynamic plots (see :class:`DynamicPlotManager`)
    '''

    def __init__(self, parent):
        '''Passing the parent object (the main window) is mandatory'''
        DynamicPlotManager.__init__(self, parent)

        self._createPermanentPanels()

        # connect the broker to shared data
        Qt.qApp.SDM.connectReader("doorName", self.setModel)
        Qt.qApp.SDM.connectReader("expConfChanged", self.onExpConfChanged)
        Qt.qApp.SDM.connectWriter("shortMessage", self, 'newShortMessage')

    def setModel(self, doorname):
        ''' Reimplemented from :class:`DynamicPlotManager`.'''
        # disconnect the previous door
        door = self.getModelObj()
        if door is not None:  # disconnect it from *all* shared data providing
            SDM = Qt.qApp.SDM
            SDM.disconnectWriter("macroStatus", door, "macroStatusUpdated")
            SDM.disconnectWriter("doorOutputChanged", door, "outputUpdated")
            SDM.disconnectWriter("doorInfoChanged", door, "infoUpdated")
            SDM.disconnectWriter("doorWarningChanged", door, "warningUpdated")
            SDM.disconnectWriter("doorErrorChanged", door, "errorUpdated")
            SDM.disconnectWriter("doorDebugChanged", door, "debugUpdated")
            SDM.disconnectWriter("doorResultChanged", door, "resultUpdated")
            SDM.disconnectWriter("expConfChanged", door,
                                 "experimentConfigurationChanged")

        # set the model
        DynamicPlotManager.setModel(self, doorname)

        # connect the new door
        door = self.getModelObj()
        if door is not None:
            SDM = Qt.qApp.SDM
            SDM.connectWriter("macroStatus", door, "macroStatusUpdated")
            SDM.connectWriter("doorOutputChanged", door, "outputUpdated")
            SDM.connectWriter("doorInfoChanged", door, "infoUpdated")
            SDM.connectWriter("doorWarningChanged", door, "warningUpdated")
            SDM.connectWriter("doorErrorChanged", door, "errorUpdated")
            SDM.connectWriter("doorDebugChanged", door, "debugUpdated")
            SDM.connectWriter("doorResultChanged", door, "resultUpdated")
            SDM.connectWriter("expConfChanged", door,
                              "experimentConfigurationChanged")

    def _createPermanentPanels(self):
        '''creates panels on the main window'''
        from sardana.taurus.qt.qtgui.extra_macroexecutor import \
            TaurusMacroExecutorWidget, TaurusSequencerWidget, \
            TaurusMacroConfigurationDialog, TaurusMacroDescriptionViewer, \
            DoorOutput, DoorDebug, DoorResult

        from sardana.taurus.qt.qtgui.extra_sardana import \
            ExpDescriptionEditor, SardanaEditor

        mainwindow = self.parent()

        # Create macroconfiguration dialog & action
        self.__macroConfigurationDialog = \
            TaurusMacroConfigurationDialog(mainwindow)
        self.macroConfigurationAction = mainwindow.taurusMenu.addAction(
            Qt.QIcon.fromTheme("preferences-system-session"),
            "Macro execution configuration...",
            self.__macroConfigurationDialog.show)

        SDM = Qt.qApp.SDM
        SDM.connectReader("macroserverName",
                          self.__macroConfigurationDialog.selectMacroServer)
        SDM.connectReader("doorName",
                          self.__macroConfigurationDialog.selectDoor)
        SDM.connectWriter("macroserverName", self.__macroConfigurationDialog,
                          'macroserverNameChanged')
        SDM.connectWriter("doorName", self.__macroConfigurationDialog,
                          'doorNameChanged')

        # Create ExpDescriptionEditor dialog
        self.__expDescriptionEditor = ExpDescriptionEditor(plotsButton=False)
        SDM.connectReader("doorName", self.__expDescriptionEditor.setModel)
        mainwindow.createPanel(self.__expDescriptionEditor,
                               'Experiment Config',
                               registerconfig=True,
                               icon=Qt.QIcon.fromTheme('preferences-system'),
                               permanent=True)
        ###############################
        # TODO: These lines can be removed once the door does emit
        # "experimentConfigurationChanged" signals
        SDM.connectWriter("expConfChanged", self.__expDescriptionEditor,
                          "experimentConfigurationChanged")
        ################################

        # put a Macro Executor
        self.__macroExecutor = TaurusMacroExecutorWidget()
        SDM.connectReader("macroserverName", self.__macroExecutor.setModel)
        SDM.connectReader("doorName", self.__macroExecutor.onDoorChanged)
        SDM.connectReader("macroStatus",
                          self.__macroExecutor.onMacroStatusUpdated)
        SDM.connectWriter("macroName", self.__macroExecutor,
                          "macroNameChanged")
        SDM.connectWriter("executionStarted", self.__macroExecutor,
                          "macroStarted")
        SDM.connectWriter("plotablesFilter", self.__macroExecutor,
                          "plotablesFilterChanged")
        SDM.connectWriter("shortMessage", self.__macroExecutor,
                          "shortMessageEmitted")
        mainwindow.createPanel(self.__macroExecutor, 'Macros',
                               registerconfig=True, permanent=True)

        # put a Sequencer
        self.__sequencer = TaurusSequencerWidget()
        SDM.connectReader("macroserverName", self.__sequencer.setModel)
        SDM.connectReader("doorName", self.__sequencer.onDoorChanged)
        SDM.connectReader("macroStatus", self.__sequencer.onMacroStatusUpdated)
        SDM.connectWriter("macroName", self.__sequencer.tree,
                          "macroNameChanged")
        SDM.connectWriter("macroName", self.__sequencer,
                          "macroNameChanged")
        SDM.connectWriter("executionStarted", self.__sequencer,
                          "macroStarted")
        SDM.connectWriter("plotablesFilter", self.__sequencer,
                          "plotablesFilterChanged")
        SDM.connectWriter("shortMessage", self.__sequencer,
                          "shortMessageEmitted")
        mainwindow.createPanel(self.__sequencer, 'Sequences',
                               registerconfig=True, permanent=True)

        # puts a macrodescriptionviewer
        self.__macroDescriptionViewer = TaurusMacroDescriptionViewer()
        SDM.connectReader("macroserverName",
                          self.__macroDescriptionViewer.setModel)
        SDM.connectReader("macroName",
                          self.__macroDescriptionViewer.onMacroNameChanged)
        mainwindow.createPanel(self.__macroDescriptionViewer,
                               'MacroDescription', registerconfig=True,
                               permanent=True)

        # puts a doorOutput
        self.__doorOutput = DoorOutput()
        SDM.connectReader("doorOutputChanged",
                          self.__doorOutput.onDoorOutputChanged)
        SDM.connectReader("doorInfoChanged",
                          self.__doorOutput.onDoorInfoChanged)
        SDM.connectReader("doorWarningChanged",
                          self.__doorOutput.onDoorWarningChanged)
        SDM.connectReader("doorErrorChanged",
                          self.__doorOutput.onDoorErrorChanged)
        mainwindow.createPanel(self.__doorOutput, 'DoorOutput',
                               registerconfig=False, permanent=True)

        # puts doorDebug
        self.__doorDebug = DoorDebug()
        SDM.connectReader("doorDebugChanged",
                          self.__doorDebug.onDoorDebugChanged)
        mainwindow.createPanel(self.__doorDebug, 'DoorDebug',
                               registerconfig=False, permanent=True)

        # puts doorResult
        self.__doorResult = DoorResult(mainwindow)
        SDM.connectReader("doorResultChanged",
                          self.__doorResult.onDoorResultChanged)
        mainwindow.createPanel(self.__doorResult, 'DoorResult',
                               registerconfig=False, permanent=True)

        # puts sardanaEditor
        # self.__sardanaEditor = SardanaEditor()
        # SDM.connectReader("macroserverName", self.__sardanaEditor.setModel)
        # mainwindow.createPanel(self.__sardanaEditor, 'SardanaEditor',
        #                        registerconfig=False, permanent=True)

        # add panic button for aborting the door
        text = "Panic Button: stops the pool (double-click for abort)"
        self.doorAbortAction = mainwindow.jorgsBar.addAction(
            Qt.QIcon("actions:process-stop.svg"),
            text, self.__onDoorAbort)

        # store beginning of times as a datetime
        self.__lastAbortTime = datetime.datetime(1, 1, 1)

        # store doubleclick interval as a timedelta
        td = datetime.timedelta(0, 0, 1000 * Qt.qApp.doubleClickInterval())
        self.__doubleclickInterval = td

    def __onDoorAbort(self):
        '''slot to be called when the abort action is triggered.
        It sends stop command to the pools (or abort if the action
        has been triggered twice in less than self.__doubleclickInterval

        .. note:: An abort command is always preceded by an stop command
        '''
        # decide whether to send stop or abort
        now = datetime.datetime.now()
        if now - self.__lastAbortTime < self.__doubleclickInterval:
            cmd = 'abort'
        else:
            cmd = 'stop'

        door = self.getModelObj()

        # abort the door
        door.command_inout('abort')
        # send stop/abort to all pools
        pools = door.macro_server.getElementsOfType('Pool')
        for pool in pools.values():
            self.info('Sending %s command to %s' % (cmd, pool.getFullName()))
            try:
                pool.getObj().command_inout(cmd)
            except:
                self.info('%s command failed on %s', cmd, pool.getFullName(),
                          exc_info=1)
        self.newShortMessage.emit("%s command sent to all pools" %
                  cmd)
        self.__lastAbortTime = now

    def createPanel(self, widget, name, **kwargs):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.createPanel(widget, name, **kwargs)

    def getPanelWidget(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.getPanel(name).widget()

    def removePanel(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        mainwindow.removePanel(name)

    def removeTemporaryPanels(self, names=None):
        '''Remove temporary panels managed by this widget'''
        # for now, the only temporary panels are the plots
        DynamicPlotManager.removePanels(self, names=names)


if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication()

    b = DynamicPlotManager(None)

    b.setModel('door/cp1/1')

    print '...'
    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
event.py:
"""

__all__ = ["BoundMethodWeakref", "CallableRef", "EventGenerator",
           "ConfigEventGenerator", "ListEventGenerator", "EventListener",
           "AttributeEventWait", "AttributeEventIterator"]

__docformat__ = "restructuredtext"

import sys
import weakref
import threading
import time
import operator

import taurus.core


class BoundMethodWeakref(object):
    """This class represents a weak reference to a method of an object since
    weak references to methods don't work by themselves"""

    def __init__(self, bound_method, del_cb=None):
        cb = (del_cb and self._deleted)
        self.func_ref = weakref.ref(bound_method.im_func, cb)
        self.obj_ref = weakref.ref(bound_method.im_self, cb)
        if cb:
            self.del_cb = CallableRef(del_cb)
        self.already_deleted = 0

    def _deleted(self, obj):
        if not self.already_deleted:
            del_cb = self.del_cb()
            if del_cb is not None:
                del_cb(self)
                self.already_deleted = 1

    def __call__(self):
        obj = self.obj_ref()
        if obj is not None:
            func = self.func_ref()
            if func is not None:
                return func.__get__(obj)

    def __hash__(self):
        return id(self)

    def __cmp__(self, other):
        if other.__class__ == self.__class__:
            ret = cmp((self.func_ref, self.obj_ref),
                      (other.func_ref, other.obj_ref))
            return ret
        return 1

    def __repr__(self):
        obj, func = self.obj_ref(), self.func_ref()
        return 'BoundMethodWeakRef of %s.%s' % (obj, func)


def CallableRef(object, del_cb=None):
    """This function returns a callable weak reference to a callable object.
    Object can be a callable object, a function or a method.

    :param object: a callable object
    :type object: callable object
    :param del_cb: calback function. Default is None meaning to callback.
    :type del_cb: callable object or None

    :return: a weak reference for the given callable
    :rtype: BoundMethodWeakref or weakref.ref"""
    if hasattr(object, 'im_self'):
        if object.im_self is not None:
            return BoundMethodWeakref(object, del_cb)
    return weakref.ref(object, del_cb)


class EventStack(object):
    "internal usage event stack"

    def __init__(self, history=True):
        self.unread_stack = []
        self.read_stack = []
        self.history = history

    def isEmpty(self):
        return len(self.unread_stack) == 0

    def push(self, event):
        self.unread_stack.append(event)

    def getNext(self):
        if self.isEmpty():
            return None
        event = self.unread_stack[0]
        del self.unread_stack[0]
        if self.history:
            self.read_stack.append(event)

        return event

    def getAllUnread(self):
        unread = self.unread_stack
        self.unread_stack = []
        return unread

    def getAllRead(self):
        read = self.read_stack
        self.read_stack = []
        return read


from object import Object


class EventGenerator(Object):
    """Base class capable of firing events"""

    WaitTimeout = 0.1

    def __init__(self, name, events_active=True):
        """Event generator constructor.

        :param name: the event generator name
        :type  name: str
        :param events_active: generate events (default is True)
        :type  events_active: bool
        :return: a new EventGenerator
        :rtype: EventGenerator"""
        self.call__init__(Object)
        self.event_name = name
        self.cb_list = []
        self.last_val = None
        self.first_event_val = None
        self.events_active = events_active
        self.cond = threading.Condition()
        self.wait_list = []

    def lock(self):
        """Locks this event generator"""
        self.cond.acquire()

    def unlock(self):
        """Unlocks this event generator"""
        if self.cond._is_owned():
            self.cond.release()
        else:
            pass

    def subscribeEvent(self, cb, data=None):
        """Subscribes to the event

        :param cb: a callable object
        :type  cb: callable
        :param data: extra data to send each time an event is triggered on the
                     given callback. Default is None
        :type  data: object"""
        if not self.events_active:
            raise RuntimeError, ('%s does not have '
                                 'events/polling active' % self.event_name)

        cb_ref = CallableRef(cb, self.unsubscribeDeletedEvent)

        try:
            self.lock()
            if (cb_ref, data) in self.cb_list:
                raise RuntimeError, ('Callback %s(%s) already reg. on %s' %
                                     (cb, data, self.event_name))
            self.cb_list.append((cb_ref, data))
            cb(data, self.first_event_val)
        finally:
            self.unlock()

    def unsubscribeDeletedEvent(self, cb_ref):
        """for internal usage only"""
        try:
            self.lock()
            aux_list = list(self.cb_list)
            for i in xrange(len(aux_list) - 1, -1, -1):
                pair = self.cb_list[i]
                if pair[0] is cb_ref:
                    del self.cb_list[i]
        finally:
            self.unlock()

    def unsubscribeEvent(self, cb, data=None):
        """Unsubscribes the given callback from the event. If the callback is not
        a listener for this event a debug message is generated an nothing happens.

        :param cb: a callable object
        :type  cb: callable
        :param data: extra data to send each time an event is triggered on the
                     given callback. Default is None
        :type  data: object"""
        cb_ref = CallableRef(cb)
        try:
            self.lock()
            if (cb_ref, data) in self.cb_list:
                self.cb_list.remove((cb_ref, data))
            else:
                self.debug("Trying to unsubscribe: %s is not a listener of %s" % (
                    str(cb_ref), self.event_name))
        finally:
            self.unlock()

    def isSubscribed(self, cb, data=None):
        """Determines if the given callback is registered for this event.

        :param cb: a callable object
        :type  cb: callable
        :param data: extra data to send each time an event is triggered on the
                     given callback. Default is None
        :type  data: object
        :return: True if callback is registered or False otherwise
        :rtype: bool"""
        cb_ref = CallableRef(cb, self.unsubscribeDeletedEvent)
        return (cb_ref, data) in self.cb_list

    def setEventsActive(self, events_active):
        """(De)activates events on this event generator.

        :param events_active: activate/deactivate events
        :type  events_active: bool"""
        self.events_active = events_active

    def getEventsActive(self):
        """Determines is events are active
        :return: True if events are active or False otherwise
        :rtype: bool"""
        return self.events_active

    def fireEvent(self, val, event_val=None):
        """Fires an event.
        :param val: event value
        :type  val: object"""
        try:
            self.lock()
            self.last_val = val
            self.cond.notifyAll()
            if event_val is None:
                event_val = val
            self.first_event_val = event_val
            for stack in self.wait_list:
                stack.push(event_val)
            for cb_ref, data in self.cb_list:
                cb = cb_ref()
                if cb is not None:
                    cb(data, event_val)
        finally:
            self.unlock()

    def waitEvent(self, val=None, equal=True, any=False, timeout=None,
                  stack=None):
        """Waits for an event to occur

        :param val: event value
        :type  val: object
        :param equal: check for equality. Default is True
        :type  equal: bool
        :param any: if True unblock after first event, not matter what value it
                    has. Default is False.
        :type  any: bool
        :param timeout: maximum time to wait (seconds). Default is None meaning
               wait forever.
        :type  timeout: float
        :param stack: For internal usage only.
        :return: the value of the event that unblocked the wait
        :rtype: object"""
        if not self.events_active:
            raise RuntimeError, ('%s does not have '
                                 'events/polling active' % self.event_name)
        try:
            self.lock()
            t0 = time.time()
            timeout_expired = False
            stack = stack or EventStack(history=False)
            while True:
                curr_val = self.last_val
                avail = not stack.isEmpty()
                if avail:
                    curr_val = stack.getNext()
                if any:
                    block = not avail
                else:
                    block = ((equal and (val != curr_val)) or
                             (not equal and (val == curr_val)))
                if timeout:
                    timeout_expired = (time.time() - t0 > timeout)
                if not block or timeout_expired:
                    break
                self.wait_list.append(stack)
                self.cond.wait(self.WaitTimeout)
                self.wait_list.remove(stack)
            val = curr_val
        finally:
            self.unlock()

        return val

    def read(self):
        """Read the last event

        :return: the last event value
        :rtype: object"""
        return self.last_val


class EventListener(object):
    """A class that listens for an event with a specific value

    Note: Since this class stores for each event value the last timestamp when
    it occured, it should only be used for events for which the event value
    domain (possible values) is limited and well known (ex: an enum)"""

    def __init__(self):
        self.last_val = None
        self.cond = threading.Condition()

        # a set implemented as a dictionary
        # dict<object, float>
        # key - event value
        # value - timestamp of last event with that value
        self.event_set = {}
        self.attr.addListener(self)

    def lock(self):
        """Locks this event listener"""
        self.cond.acquire()

    def unlock(self):
        """Unlocks this event listener"""
        if self.cond._is_owned():
            self.cond.release()
        else:
            pass

    def clearEventSet(self):
        "Clears the internal event buffer"
        self.event_set.clear()

    def fireEvent(self, v):
        """Notifies that a given event has arrived
        This function is protected inside with the object's lock. Do NOT call
        this function when you have the lock acquired on this object.

        :param v: event value
        :type  v: object"""
        try:
            t = time.time()
            self.lock()
            self.last_val = v
            self.event_set[v] = t
            self.cond.notifyAll()
        finally:
            self.unlock()

    def waitEvent(self, val, after=0, equal=True):
        """Wait for an event with the given value. You MUST protect this function
        with this object's lock before calling this method and always unlock it
        afterward, of course::

            from taurus.core.util.event import EventListener

            class MyEvtListener(EventListener):
                # Your specific listener code here
                pass

            evt_listener = EventListener()
            try:
                evt_listener.lock()
                t = time.time()
                go()
                evt_listener.waitEvent(Stop, t)
            finally:
                evt_listener.unlock()

        :param val: value to compare
        :type  val: object
        :param after: timestamp. wait for events comming after the given time.
                      default value is 0 meaning any event after Jan 1, 1970
        :type  after: float
        :param equal: compare for equality. equal=True means an event with the
                      given value, equal=False means any event which as a different value
        :type  equal: bool
        """
        s = self.event_set
        while True:
            if equal:
                t = s.get(val)
                if t and t >= after:
                    return
            else:
                for v, t in s.items():
                    if v == val:
                        continue
                    if t >= after:
                        return
            self.cond.wait()


class ConfigEventGenerator(EventGenerator):
    """Manage configuration events"""

    def fireEvent(self, val, event_val=None):
        EventGenerator.fireEvent(self, val, event_val)


class ListEventGenerator(EventGenerator):
    """Manage list events, detecting changes in the list"""

    def __init__(self, name, events_active=True):
        self.call__init__(EventGenerator, name, events_active)
        self.last_val = []
        self.first_event_val = [], [], []

    def fireEvent(self, val):
        """Notifies that a given event has arrived
        This function is protected inside with the object's lock. Do NOT call
        this function when you have the lock acquired on this object.

        :param val: event value
        :type  val: object"""
        # if attribute is not alive and last time was also not alive then
        # skip the event propagation
        if val is None:
            if len(self.last_val) == 0:
                return
            val = []

        try:
            self.lock()
            val = list(val)
            last_val = self.last_val
            rm = [x for x in last_val if x not in val]
            add = [x for x in val if x not in last_val]
            event_val = val, rm, add
            EventGenerator.fireEvent(self, val, event_val)
            self.first_event_val = val, [], val
        finally:
            self.unlock()


class AttributeEventWait(object):
    """Class designed to connect to a :class:`taurus.core.taurusattribute.TaurusAttribute` and
    fire events or wait for a certain event."""

    def __init__(self, attr=None):
        self._last_val = None
        self._attr = None
        self._cond = threading.Condition()
        self._event_set = {}
        if attr is not None:
            self.connect(attr)

    def connect(self, attr):
        """Connect to the given attribute
        :param attr: the attribute to connect to
        :type  attr: taurus.core.taurusattribute.TaurusAttribute"""
        needAdd = True
        if self._attr is not None:
            if attr == self._attr:
                needAdd = False
            else:
                self._attr.removeListener(self)
        self.clearEventSet()
        self._last_val = None
        self._attr = attr
        if needAdd:
            self._attr.addListener(self)

    def disconnect(self):
        """Disconnects from the attribute. If not connected nothing happens."""
        self.clearEventSet()
        if self._attr:
            self._attr.removeListener(self)
        self._attr = None
        self._last_val = None

    def lock(self):
        """Locks this event listener"""
        self._cond.acquire()

    def unlock(self):
        """Unocks this event listener"""
        if self._cond._is_owned():
            self._cond.release()
        else:
            lock = getattr(self._cond, "_Condition__lock")
            th = getattr(lock, "_RLock__owner")
            curr_th = threading.current_thread()
            if th is not None:
                name = th.name
            else:
                name = "<unknown>"
            print "WARNING: Thread %s trying to unlock condition previously " \
                  "locked by thread %s" % (curr_th.name, name)

    def clearEventSet(self):
        "Clears the internal event buffer"
        self._event_set.clear()
        self._last_val = None

    def eventReceived(self, s, t, v):
        """Event listener method for the underlying attribute. Do not call this
        method. It will be called internally when the attribute generates
        an event."""
        if t == taurus.core.taurusbasetypes.TaurusEventType.Config:
            return
        elif t == taurus.core.taurusbasetypes.TaurusEventType.Error:
            self.fireEvent(None)
        else:
            self.fireEvent(v.value)

    def fireEvent(self, v):
        """Notifies that a given event has arrived
        This function is protected inside with the object's lock. Do NOT call
        this function when you have the lock acquired on this object.

        :param v: event value
        :type  v: object"""
        t = time.time()
        self.lock()
        try:
            self._last_val = v
            self._event_set[v] = t
            self._cond.notifyAll()
        finally:
            self.unlock()

    def getLastRecordedEvent(self):
        """returns the value of the last recorded event or None if no event has
        been received or the last event was an error event

        :return: the last event value to be recorded
        :rtype: object"""
        return self._last_val

    def getRecordedEvents(self):
        """Returns a reference to the internal dictionary used to store the internal
        events. Modify the return dictionary at your own risk!

        :return: reference to the internal event dictionary
        :rtype: dict"""
        return self._event_set

    def getRecordedEvent(self, v):
        """Returns the the recorded local timestamp for the event with the given
        value or None if no event with the given value has been recorded.

        :param v: event value
        :type  v: object
        :return: local timestamp for the event or None if no event has been recorded
        :rtype: float"""
        return self._event_set.get(v)

    def waitEvent(self, val, after=0, equal=True, timeout=None, retries=-1,
                  any=False):
        """Wait for an event with the given value.

        :param val: value to compare
        :type  val: object
        :param after: timestamp. wait for events comming after the given time.
                      default value is 0 meaning any event after Jan 1, 1970
        :type  after: float
        :param equal: compare for equality. equal=True means an event with the
                      given value, equal=False means any event which as a different value
        :type  equal: bool
        :param timeout: maximum time to wait (seconds). Default is None meaning
               wait forever.
        :type  timeout: float
        :param retries: number of maximum retries of max timeout to attempts.
                        Default is -1 meaning infinite number of retries.
                        0 means no wait. Positive number is obvious.
        :param any: if any is True ignore 'val' parameter and accept any event.
                    If False (default),check with given 'val' parameter
        :type  any: bool
        """
        if retries == 0:
            return
        if timeout is None:
            # if waitting forever doesn't make sense to retry
            retries = 1
        if after is None:
            after = 0
        s = self._event_set
        self.lock()
        try:
            # increase the retries by one just because of how the loop is done
            if retries > 0:
                retries += 1
            while retries != 0:
                if any:
                    for v, t in s.items():
                        if t >= after:
                            return
                if equal:
                    t = s.get(val)
                    if (t is not None) and (t >= after):
                        return
                else:
                    for v, t in s.items():
                        if v == val:
                            continue
                        if t >= after:
                            return
                self._cond.wait(timeout)
                retries -= 1
        except Exception, e:
            sys.stderr.write(
                "AttributeEventWait: Caught exception while waitting: %s\n" % str(e))
            raise e
        finally:
            self.unlock()


class AttributeEventIterator(object):
    """Internal class. For test purposes"""

    def __init__(self, *attrs):
        self._attrs = None
        self._cond = threading.Condition()
        if len(attrs) > 0:
            self.connect(attrs)

    def connect(self, attrs):
        if not operator.isSequenceType(attrs):
            attrs = (attrs,)
        self.disconnect()
        self._attrs = attrs
        for attr in self._attrs:
            attr.addListener(self)

    def disconnect(self):
        if self._attrs is None:
            return
        for attr in self._attrs:
            attr.removeListener(self)

    def lock(self):
        self._cond.acquire()

    def unlock(self):
        if self._cond._is_owned():
            self._cond.release()
        else:
            lock = getattr(self._cond, "_Condition__lock")
            th = getattr(lock, "_RLock__owner")
            curr_th = threading.current_thread()
            print "WARNING: Thread %s trying to unlock condition previously " \
                  "locked by thread %s" % (curr_th.name, th.name)

    def eventReceived(self, s, t, v):
        if t not in (taurus.core.taurusbasetypes.TaurusEventType.Change, taurus.core.taurusbasetypes.TaurusEventType.Periodic):
            return
        self.fireEvent(s, v.value)

    def fireEvent(self, s, v):
        t = time.time()
        self.lock()
        try:
            self._data = s, v
            self._cond.notifyAll()
        finally:
            self.unlock()

    def events(self, timeout=1):
        self.lock()
        try:
            while True:
                self._cond.wait(timeout)
                yield self._data
        except Exception, e:
            print "INFO: Caught exception while waiting", str(e)
        finally:
            self.unlock()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the set of base class taurus controllers."""

__all__ = ["TaurusBaseController", "TaurusAttributeControllerHelper",
           "TaurusScalarAttributeControllerHelper",
           "TaurusConfigurationControllerHelper",
           "updateLabelBackground"]

__docformat__ = 'restructuredtext'

import weakref

from taurus.external.qt import Qt

from taurus.core.taurusbasetypes import DataFormat, TaurusEventType

from taurus.qt.qtgui.util import QT_ATTRIBUTE_QUALITY_PALETTE
from taurus.qt.qtgui.util import QT_DEVICE_STATE_PALETTE


class TaurusBaseController(object):
    """Base class for all taurus controllers"""

    def __init__(self, widget, updateAsPalette=True):
        self._widget = weakref.ref(widget)
        self._updateAsPalette = updateAsPalette
        self._stateObj = None
        self._last_value = None
        self._last_config_value = None
        self._last_error_value = None
        self._setStyle()

    def _setStyle(self):
        pass

    def usePalette(self):
        return self._updateAsPalette

    def widget(self):
        return self._widget()

    def modelObj(self):
        return self.widget().getModelObj()

    attrObj = configObj = deviceObj = modelObj

    def valueObj(self):
        value = self._last_value
        if value is None:
            modelObj = self.modelObj()
            if modelObj is None:
                return None
            value = modelObj.getValueObj()
        return value

    def value(self):
        valueObj = self.valueObj()
        return getattr(valueObj, "value", None)

    def w_value(self):  # TODO: adapt to tep14
        valueObj = self.valueObj()
        return getattr(valueObj, "w_value", None)

    def quality(self):
        valueObj = self.valueObj()
        return getattr(valueObj, "quality", None)

    def state(self):
        return self._stateObj.state

    def getDisplayValue(self, write=False):
        return self.widget().getDisplayValue()

    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_src == self.modelObj():  # update the "_last" values only if the event source is the model (it could be the background...)
            if evt_type == TaurusEventType.Change or evt_type == TaurusEventType.Periodic:
                self._last_value = evt_value
            elif evt_type == TaurusEventType.Config:  # TODO: adapt to tep14
                self._last_config_value = evt_value
            else:
                self._last_error_value = evt_value
                # In case of error, modify the last_value as well
                try:
                    self._last_value = self.modelObj().getValueObj()
                except:
                    self._last_value = None
        self.update()

    def eventReceived(self, evt_src, evt_type, evt_value):
        # should handle the state event here. Because this is invoked by a random
        # thread, we pass it to the widget, which will forward to the proper
        # thread

        #@todo: sometimes we get this method called but self.widget() is None. Check why.
        #       For the moment I just protect it by substituting the following line by the ones after it
        #self.widget().eventReceived(evt_src, evt_type, evt_value)
        w = self.widget()
        if w is not None:
            w.eventReceived(evt_src, evt_type, evt_value)

    def update(self):
        widget = self.widget()
        self._updateConnections(widget)
        self._updateForeground(widget)
        self._updateBackground(widget)
        self._updateToolTip(widget)

    def _needsStateConnection(self):
        return False

    def _updateConnections(self, widget):
        stateObj, newStateObj = self._stateObj, None
        if self._needsStateConnection():
            newStateObj = self.deviceObj()
            if stateObj != newStateObj:
                if stateObj is not None:
                    stateObj.removeListener(self)
                if newStateObj is not None:
                    newStateObj.addListener(self)
        else:
            if stateObj is not None:
                stateObj.removeListener(self)
        self._stateObj = newStateObj

    def _updateForeground(self, widget):
        pass

    def _updateBackground(self, widget):
        pass

    def _updateToolTip(self, widget):
        if widget.getAutoTooltip():
            widget.setToolTip(widget.getFormatedToolTip())


class TaurusAttributeControllerHelper(object):

    def configObj(self):
        return self.attrObj()  # they are the same object since tep14
        # attrObj = self.attrObj()
        # if attrObj is None: return None
        # return attrObj.getConfig()

    def deviceObj(self):
        attrObj = self.attrObj()
        if attrObj is None:
            return None
        return attrObj.getParentObj()


class TaurusScalarAttributeControllerHelper(TaurusAttributeControllerHelper):

    def getDisplayValue(self, write=False):
        valueObj = self.valueObj()
        widget = self.widget()
        if valueObj is None or valueObj.rvalue is None:
            return widget.getDisplayValue()

        format = self.attrObj().data_format
        if format == DataFormat._0D:
            return self._getDisplayValue(widget, valueObj, None, write)

        idx = widget.getModelIndexValue()
        return self._getDisplayValue(widget, valueObj, idx, write)

    def _getDisplayValue(self, widget, valueObj, idx, write):
        try:
            if write:
                value = valueObj.wvalue
            else:
                value = valueObj.rvalue
            if idx is not None and len(idx):
                for i in idx:
                    value = value[i]
            if self._last_config_value is None or value is None:
                return widget.displayValue(value)
        except Exception, e:
            return widget.getNoneValue()

    def displayValue(self, value):
        if value is None:
            return None
        ret = None
        try:
            if self.isScalar():
                format = self.getFormat()
                if self.isNumeric() and format is not None:
                    format = self.getFormat()
                    ret = self.getFormat() % value
                else:
                    ret = str(value)
            elif self.isSpectrum():
                ret = str(value)
            else:
                ret = str(value)
        except:
            # if cannot calculate value based on the format just return the
            # value
            raise
            ret = str(value)
        return ret


class TaurusConfigurationControllerHelper(object):
    #TODO: Check if this is  used. If so, rename to avoid "Configuration"

    def __init__(self):
        self._configParam = None

    def attrObj(self):
        return self.configObj()  # they are the same object since tep14

    def deviceObj(self):
        attrObj = self.attrObj()
        if attrObj is None:
            return None
        return attrObj.getParentObj()

    @property
    def configParam(self):
        if self._configParam is None:
            self._configParam = self.widget().getModelFragmentName() or ''
        return self._configParam

    def getDisplayValue(self, write=False):
        widget = self.widget()
        model = self.configObj()
        if model is None:
            return widget.getNoneValue()
        param = self.configParam
        try:
            val = widget.getModelFragmentObj()
            try:
                no_val = getattr(model, "no_" + param)  # TODO: Tango-centric
                if val.lower() == no_val.lower():
                    val = widget.getNoneValue()
            except:
                pass
        except AttributeError:
            if param:
                val = str(param)
                attr = self.attrObj()
                if attr is not None:
                    val = val.replace('<label>', attr.label or '---')
                    val = val.replace('<attr_name>', attr.name or '---')
                    val = val.replace('<attr_fullname>', attr.getFullName() or
                                      '---')
                dev = self.deviceObj()
                if dev is not None:
                    val = val.replace('<dev_fullname>', dev.getFullName() or
                                      '---')
                    val = val.replace('<dev_alias>', dev.getSimpleName() or
                                      '---')
                    val = val.replace('<dev_name>', dev.getNormalName() or
                                      '---')
            else:
                val = widget.getNoneValue()
        except:
            widget.debug("Invalid configuration parameter '%s'" % param)
            val = widget.getNoneValue()
        if val is None:
            val = widget.getNoneValue()
        return val


StyleSheetTemplate = """border-style: outset; border-width: 2px; border-color: {0}; {1}"""


def _updatePaletteColors(widget, bgBrush, fgBrush, frameBrush):
    qt_palette = widget.palette()
    qt_palette.setBrush(Qt.QPalette.Window, bgBrush)
    qt_palette.setBrush(Qt.QPalette.Base, bgBrush)
    qt_palette.setBrush(Qt.QPalette.WindowText, fgBrush)
    qt_palette.setBrush(Qt.QPalette.Light, frameBrush)
    qt_palette.setBrush(Qt.QPalette.Dark, frameBrush)
    widget.setPalette(qt_palette)


def updateLabelBackground(ctrl, widget):
    """Helper method to setup background of taurus labels and lcds"""
    bgRole = widget.bgRole

    if ctrl.usePalette():
        widget.setAutoFillBackground(True)
        if bgRole in ('', 'none'):
            transparentBrush = Qt.QBrush(Qt.Qt.transparent)
            frameBrush = transparentBrush
            bgBrush, fgBrush = transparentBrush, Qt.QBrush(Qt.Qt.black)
        else:
            frameBrush = Qt.QBrush(Qt.QColor(255, 255, 255, 128))
            bgItem, palette = None, QT_DEVICE_STATE_PALETTE
            if bgRole == 'quality':
                palette = QT_ATTRIBUTE_QUALITY_PALETTE
                bgItem = ctrl.quality()
            elif bgRole == 'state':
                bgItem = ctrl.state()
            elif bgRole == 'value':
                bgItem = ctrl.value()
            bgBrush, fgBrush = palette.qbrush(bgItem)
        _updatePaletteColors(widget, bgBrush, fgBrush, frameBrush)
    else:
        if bgRole in ('', 'none'):
            ss = StyleSheetTemplate.format("rgba(0,0,0,0)", "")
        else:
            bgItem, palette = None, QT_DEVICE_STATE_PALETTE
            if bgRole == 'quality':
                palette = QT_ATTRIBUTE_QUALITY_PALETTE
                bgItem = ctrl.quality()
            elif bgRole == 'state':
                bgItem = ctrl.state()
            elif bgRole == 'value':
                bgItem = ctrl.value()
            color_ss = palette.qtStyleSheet(bgItem)
            ss = StyleSheetTemplate.format("rgba(255,255,255,128)",  color_ss)
        widget.setStyleSheet(ss)
    widget.update()  # necessary in pyqt <= 4.4



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This package contains a collection of taurus console widgets"""

__docformat__ = 'restructuredtext'

import weakref

from taurus import Device
from taurus.core.util.enumeration import Enumeration
from taurus.external.qt import Qt

from IPython.config.loader import Config


EnvironmentMode = Enumeration("EnvironmentMode", ("Overwrite", "Merge"))


class BaseConsoleExtension(object):

    Name = 'ipython'
    Label = 'IPython'

    def __init__(self, console_factory, profile='default', config=None,
                 mode=EnvironmentMode.Merge):
        self.console_factory = console_factory
        self.profile = profile
        self.profile_arg = '--profile=' + profile
        self.config = config or Config()
        self.mode = mode

    def __enter__(self):
        app = self.console_factory.get_ipython_application()
        self.orig_config = app.config
        self.orig_kernel_argv = argv = app.kernel_argv
        argv, has_profile = list(argv), False
        for i, arg in enumerate(argv):
            if arg.startswith('--profile'):
                has_profile = True
                if self.mode == EnvironmentMode.Overwrite:
                    argv[i] = self.profile_arg
        if not has_profile:
            argv.append(self.profile_arg)
        app.kernel_argv = argv

        config = self.config
        if self.mode == EnvironmentMode.Merge:
            config = config.copy()
            config.update(app.config)
        app.config = config

    def __exit__(self, exc_type, exc_value, traceback):
        app = self.console_factory.get_ipython_application()
        app.config = self.orig_config
        app.kernel_argv = self.orig_kernel_argv

    @classmethod
    def is_enabled(cls):
        return True


class TangoConsoleExtension(BaseConsoleExtension):

    Name = 'tango'
    Label = 'Tango'

    def __init__(self, console_factory, config=None):
        if config is None:
            config = Config()
        import PyTango.ipython
        PyTango.ipython.load_config(config)
        super(TangoConsoleExtension, self).__init__(console_factory,
                                                    profile='tango',
                                                    config=config)

    @classmethod
    def is_enabled(cls):
        try:
            import PyTango
            v = list(PyTango.__version_info__[:2])
            return v >= [7, 2]
        except:
            return False


from IPython.core.profiledir import ProfileDir, ProfileDirError
from IPython.utils.path import get_ipython_dir


def create_sardana_profile(profile, door_name):

    ipython_dir = get_ipython_dir()
    try:
        ProfileDir.find_profile_dir_by_name(ipython_dir, profile)
    except ProfileDirError:
        from sardana.spock.genutils import create_spock_profile
        create_spock_profile(ipython_dir, "spock", profile, door_name)


def get_profile_from_args(args):
    for arg in args:
        if arg.startswith("--profile="):
            profile = arg[10:]
            return True, profile
    return False, "spockdoor"


class SDMDoorReader(Qt.QObject):

    def __init__(self, console, sdm=None, parent=None):
        super(Qt.QObject, self).__init__(parent)
        self._console = weakref.ref(console)
        if sdm is None:
            if not hasattr(Qt.qApp, 'SDM'):
                raise Exception("Cannot connect to shared data manager")
            sdm = Qt.qApp.SDM
        sdm.connectReader("doorName", self.onDoorChanged)

    @property
    def console(self):
        return self._console()

    def onDoorChanged(self, door_name):
        door = Device(door_name)
        dalias, dname = door.getSimpleName(), door.getNormalName()
        create_sardana_profile(dalias, dname)


class SardanaConsoleExtension(BaseConsoleExtension):

    Name = 'spock'
    Label = 'Spock'

    def fill_sardana_config(self, config):
        import sardana.spock
        profile = 'spockdoor'
        if not config.Spock:
            if hasattr(Qt.qApp, 'SDM'):
                dm = Qt.qApp.SDM.getDataModelProxy('doorName')
                if dm is not None:
                    door_name = dm.getData()
                    door = Device(door_name)
                    dalias, dname = door.getSimpleName(), door.getNormalName()
                    create_sardana_profile(dalias, dname)
                    profile = dalias
                    config.Spock.door_name = dname
        sardana.spock.load_config(config)
        return profile

    def __init__(self, console_factory, config=None):
        if config is None:
            config = Config()
        profile = self.fill_sardana_config(config)
        super(SardanaConsoleExtension, self).__init__(console_factory,
                                                      profile=profile,
                                                      config=config)

    @classmethod
    def is_enabled(cls):
        try:
            import sardana
            v = list(sardana.Release.version_info[:2])
            return v >= [1, 2]
        except:
            return False



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.panel import TaurusForm

app = Qt.QApplication(sys.argv)

panel = TaurusForm()
props = ['state', 'status', 'position', 'velocity', 'acceleration']
model = ['sys/taurustest/1/%s' % p for p in props]
panel.setModel(model)
panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a Taurus widget based on QLCDNumber"""

__all__ = ["TaurusLCD"]

__docformat__ = 'restructuredtext'

import operator

# shame of me for importing PyTango!
import PyTango

from taurus.core.taurusbasetypes import TaurusElementType, TaurusEventType
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.base import TaurusBaseController
from taurus.qt.qtgui.base import TaurusScalarAttributeControllerHelper
from taurus.qt.qtgui.base import TaurusConfigurationControllerHelper
from taurus.qt.qtgui.base import updateLabelBackground

_QT_PLUGIN_INFO = {
    'module': 'taurus.qt.qtgui.display',
    'group': 'Taurus Display',
    'icon': "designer:lcdnumber.png",
}

TaurusModelType = TaurusElementType
EventType = TaurusEventType

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Controller classes for LCD
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-


class TaurusLCDController(TaurusBaseController):

    def _setStyle(self):
        TaurusBaseController._setStyle(self)
        lcd = self.lcd()
        if self.usePalette():
            lcd.setFrameShape(Qt.QFrame.Box)
            lcd.setFrameShadow(Qt.QFrame.Raised)
            lcd.setLineWidth(1)

    def lcd(self):
        """Helper method that returns the LCDNumber widget"""
        return self.widget()

    def _updateForeground(self, lcd):
        self._updateLength(lcd)
        self._updateValue(lcd)

    def _needsStateConnection(self):
        lcd = self.lcd()
        return 'state' in (lcd.fgRole, lcd.bgRole)

    def _getCharsToDisplayFromFormat(self, fmt):
        n = ''
        for c in fmt:
            if c == '.':
                break
            if c.isdigit():
                n += c
        return int(n)

    def _updateLength(self, lcd):
        model = self.modelObj()
        n = 6
        try:
            fmt = model.getFormat()
            if fmt:
                n = self._getCharsToDisplayFromFormat(fmt)
        except:
            pass
        lcd.setNumDigits(n)

    def _updateValue(self, lcd):
        fgRole, value = lcd.fgRole, ""
        if fgRole == 'value':
            w = self.getDisplayValue()
            # @todo: sometimes getDisplayValue called but self.widget() is None.
            # For the moment it is just protected by substituting the following
            # line by the ones after it.
            # value += self.getDisplayValue()
            if w is not None:
                value += self.getDisplayValue()
        elif fgRole == 'w_value':
            value += self.getDisplayValue(True)
        elif fgRole in ('', 'none'):
            pass
        else:
            lcd.display("udef")
            return
        lcd.display(value)

    _updateBackground = updateLabelBackground


class TaurusLCDControllerAttribute(TaurusScalarAttributeControllerHelper, TaurusLCDController):

    def __init__(self, lcd):
        TaurusScalarAttributeControllerHelper.__init__(self)
        TaurusLCDController.__init__(self, lcd)

    def _getDisplayValue(self, widget, valueObj, idx, write):
        try:
            if write:
                value = valueObj.wvalue.magnitude
            else:
                value = valueObj.rvalue.magnitude
            if idx is not None and len(idx):
                for i in idx:
                    value = value[i]
            if self._last_config_value is None or value is None:
                return widget.displayValue(value)
        except Exception, e:
            return widget.getNoneValue()


class TaurusLCDControllerConfiguration(TaurusConfigurationControllerHelper, TaurusLCDController):

    def __init__(self, lcd):
        TaurusConfigurationControllerHelper.__init__(self)
        TaurusLCDController.__init__(self, lcd)


#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Design time controllers for LCD
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

class TaurusLCDControllerDesignMode(object):

    def _updateLength(self, lcd):
        lcd.setNumDigits(6)

    def getDisplayValue(self, write=False):
        v = self.w_value()
        if not write:
            v = self.value()
        return "%6.2f" % v

    def value(self):
        return 99.99

    def w_value(self):
        return 0.0

    def quality(self):
        return PyTango.AttrQuality.ATTR_VALID

    def state(self):
        return PyTango.DevState.ON

    def _updateToolTip(self, lcd):
        lcd.setToolTip("Some random value for design purposes only")


class TaurusLCDControllerAttributeDesignMode(TaurusLCDControllerDesignMode, TaurusLCDControllerAttribute):

    def __init__(self, label):
        TaurusLCDControllerDesignMode.__init__(self)
        TaurusLCDControllerAttribute.__init__(self, label)


class TaurusLCDControllerConfigurationDesignMode(TaurusLCDControllerDesignMode, TaurusLCDControllerConfiguration):

    def __init__(self, label):
        TaurusLCDControllerDesignMode.__init__(self)
        TaurusLCDControllerConfiguration.__init__(self, label)

    def getDisplayValue(self, write=False):
        return "%6.2f" % -99.99

    def _updateToolTip(self, lcd):
        lcd.setToolTip(
            "Some random configuration value for design purposes only")


_CONTROLLER_MAP = {
    None: TaurusLCDController,
    TaurusModelType.Unknown: TaurusLCDController,
    TaurusModelType.Attribute: TaurusLCDControllerAttribute,
    TaurusModelType.Configuration: TaurusLCDControllerConfiguration,
}

_DESIGNER_CONTROLLER_MAP = {
    None: TaurusLCDControllerAttributeDesignMode,
    TaurusModelType.Unknown: TaurusLCDControllerAttributeDesignMode,
    TaurusModelType.Attribute: TaurusLCDControllerAttributeDesignMode,
    TaurusModelType.Configuration: TaurusLCDControllerConfigurationDesignMode,
}


def Controller(lcd):
    ctrl_map = _CONTROLLER_MAP
    if lcd._designMode:
        ctrl_map = _DESIGNER_CONTROLLER_MAP

    model_type = lcd.getModelType()
    ctrl_klass = ctrl_map.get(model_type, TaurusLCDController)
    return ctrl_klass(lcd)


class TaurusLCD(Qt.QLCDNumber, TaurusBaseWidget):
    """ A Taurus-enabled :class:`Qt.QLCDNumber` widget. Its text can represent
    either the rvalue or wvalue *magnitude* (or nothing), and the background
    can colour-code the attribute quality or the device state (or nothing)"""

    DefaultBgRole = 'quality'
    DefaultFgRole = 'value'
    DefaultShowText = True
    DefaultModelIndex = None

    def __init__(self, parent=None, designMode=False):
        self._bgRole = self.DefaultBgRole
        self._fgRole = self.DefaultFgRole
        self._modelIndex = self.DefaultModelIndex
        self._modelIndexStr = ''
        self._controller = None
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QLCDNumber, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)

        # if we are in design mode there will be no events so we force the
        # creation of a controller object
        if self._designMode:
            self.controller().update()

    def _calculate_controller_class(self):
        ctrl_map = _CONTROLLER_MAP
        if self._designMode:
            ctrl_map = _DESIGNER_CONTROLLER_MAP

        model_type = self.getModelType()
        ctrl_klass = ctrl_map.get(model_type, TaurusLCDController)
        return ctrl_klass

    def controller(self):
        ctrl = self._controller
        # if there is a controller object and it is not the base controller...
        if ctrl is not None and not ctrl.__class__ == TaurusLCDController:
            return ctrl

        # if there is a controller object and it is still the same class...
        ctrl_klass = self._calculate_controller_class()
        if ctrl is not None and ctrl.__class__ == ctrl_klass:
            return ctrl

        self._controller = ctrl = ctrl_klass(self)
        return ctrl

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def handleEvent(self, evt_src, evt_type, evt_value):
        self.controller().handleEvent(evt_src, evt_type, evt_value)

    def isReadOnly(self):
        return True

    def setModel(self, m):
        # force to build another controller
        self._controller = None
        TaurusBaseWidget.setModel(self, m)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelIndexValue(self):
        return self._modelIndex

    def getModelIndex(self):
        return self._modelIndexStr

    def setModelIndex(self, modelIndex):
        mi = str(modelIndex)
        if len(mi) == 0:
            self._modelIndex = None
        else:
            try:
                mi_value = eval(str(mi))
            except:
                return
            if type(mi_value) == int:
                mi_value = mi_value,
            if not operator.isSequenceType(mi_value):
                return
            self._modelIndex = mi_value
        self._modelIndexStr = mi
        self.controller().update()

    def resetModelIndex(self):
        self.setModelIndex(self.DefaultModelIndex)

    def getBgRole(self):
        return self._bgRole

    def setBgRole(self, bgRole):
        self._bgRole = str(bgRole).lower()
        self.controller().update()

    def resetBgRole(self):
        self.setBgRole(self.DefaultBgRole)

    def getFgRole(self):
        return self._fgRole

    def setFgRole(self, fgRole):
        self._fgRole = str(fgRole).lower()
        self.controller().update()

    def resetFgRole(self):
        self.setFgRole(self.DefaultFgRole)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        d = TaurusBaseWidget.getQtDesignerPluginInfo()
        d.update(_QT_PLUGIN_INFO)
        return d

    #: This property holds the unique URI string representing the model name
    #: with which this widget will get its data from. The convention used for
    #: the string can be found :ref:`here <model-concept>`.
    #:
    #: In case the property :attr:`useParentModel` is set to True, the model
    #: text must start with a '/' followed by the attribute name.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getModel`
    #:     * :meth:`TaurusLCD.setModel`
    #:     * :meth:`TaurusBaseWidget.resetModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel, setModel,
                            TaurusBaseWidget.resetModel)

    #: This property holds whether or not this widget should search in the
    #: widget hierarchy for a model prefix in a parent widget.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getUseParentModel`
    #:     * :meth:`TaurusBaseWidget.setUseParentModel`
    #:     * :meth:`TaurusBaseWidget.resetUseParentModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    useParentModel = Qt.pyqtProperty("bool", TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)

    #: This property holds the index inside the model value that should be
    #: displayed
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLCD.getModelIndex`
    #:     * :meth:`TaurusLCD.setModelIndex`
    #:     * :meth:`TaurusLCD.resetModelIndex`
    #:
    #: .. seealso:: :ref:`model-concept`
    modelIndex = Qt.pyqtProperty(
        "QString", getModelIndex, setModelIndex, resetModelIndex)

    #: This property holds the foreground role.
    #: Valid values are:
    #:
    #:     #. ''/'None' - no value is displayed
    #:     #. 'value' - the value is displayed
    #:     #. 'w_value' - the write value is displayed
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLCD.getFgRole`
    #:     * :meth:`TaurusLCD.setFgRole`
    #:     * :meth:`TaurusLCD.resetFgRole`
    fgRole = Qt.pyqtProperty("QString", getFgRole, setFgRole,
                             resetFgRole, doc="foreground role")

    #: This property holds the background role.
    #: Valid values are ''/'None', 'quality', 'state'
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLCD.getBgRole`
    #:     * :meth:`TaurusLCD.setBgRole`
    #:     * :meth:`TaurusLCD.resetBgRole`
    bgRole = Qt.pyqtProperty("QString", getBgRole, setBgRole,
                             resetBgRole, doc="background role")


def demo():
    "LCD"
    import demo.tauruslcddemo
    return demo.tauruslcddemo.main()


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <full_attribute_name(s)>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus LCD demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        w = demo()
    else:
        models = map(str.lower, args)

        w = Qt.QWidget()
        layout = Qt.QGridLayout()
        w.setLayout(layout)
        for model in models:
            label = TaurusLCD()
            label.model = model
            layout.addWidget(label)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == '__main__':
    main()
